/**
 * Copyright (c) 2009 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Stack;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 */
public abstract class AbstractTermGenerator implements TermGenerator {

    protected static final String DEFAULT_LANGUAGE = "en";

    // avoid punctuation, including underscores
    // TODO: it would be nice to use non-ascii terms - cf bug 9069
    // protected static final String SPLIT_PATTERN = "[^\\pL\\pN]+";
    // meanwhile we will only use ascii terms
    protected static final String SPLIT_PATTERN = "[^A-Za-z0-9]+";

    public static final long LIMIT = 1000 * 1000;

    protected static final String XML_LANG_NAMESPACE = "http://www.w3.org/XML/1998/namespace";

    protected static final String XML_LANG_NAME = "lang";

    protected static SimpleLogger logger;

    protected Configuration config;

    protected XmlPullParser xpp;

    protected Reader reader;

    protected String language = DEFAULT_LANGUAGE;

    protected Object readerMutex = new Object();

    protected List<File> paths;

    protected Iterator<File> pathIterator;

    protected List<String> buffer = new Vector<String>();

    protected final static int REPETITION_QUEUE_SIZE = 10;

    protected LinkedList<String> repetitionQueue = new LinkedList<String>();

    protected Stack<String> elementStack = new Stack<String>();

    protected ZipInputStream zipInputStream;

    // subclass must initialize nameList
    protected static String[] nameList = null;

    protected static Map<String, String[]> stopWords = new HashMap<String, String[]>();

    protected static String[] languageStopWords = null;

    static {
        // english (en)
        String[] words = new String[] { "a", "about", "above", "across",
                "after", "afterwards", "again", "against", "all",
                "almost", "alone", "along", "already", "also",
                "although", "always", "am", "among", "amongst",
                "amoungst", "amount", "an", "and", "another", "any",
                "anyhow", "anyone", "anything", "anyway", "anywhere",
                "are", "around", "as", "at", "back", "be", "became",
                "because", "become", "becomes", "becoming", "been",
                "before", "beforehand", "behind", "being", "below",
                "beside", "besides", "between", "beyond", "bill", "both",
                "bottom", "but", "by", "call", "can", "cannot", "cant",
                "co", "computer", "con", "could", "couldnt", "cry", "de",
                "describe", "detail", "do", "does", "done", "down",
                "due", "during", "each", "eg", "eight", "either",
                "eleven", "else", "elsewhere", "empty", "enough", "etc",
                "even", "ever", "every", "everyone", "everything",
                "everywhere", "except", "few", "fifteen", "fify", "fill",
                "find", "fire", "first", "five", "for", "former",
                "formerly", "forty", "found", "four", "from", "front",
                "full", "further", "get", "give", "go", "had", "has",
                "hasnt", "have", "he", "hence", "her", "here",
                "hereafter", "hereby", "herein", "hereupon", "hers",
                "herself", "him", "himself", "his", "how", "however",
                "hundred", "i", "ie", "if", "in", "inc", "indeed",
                "interest", "into", "is", "it", "its", "itself", "keep",
                "last", "latter", "latterly", "least", "less", "ltd",
                "made", "many", "may", "me", "meanwhile", "might",
                "mill", "mine", "more", "moreover", "most", "mostly",
                "move", "much", "must", "my", "myself", "name", "namely",
                "neither", "never", "nevertheless", "next", "nine", "no",
                "nobody", "none", "noone", "nor", "not", "nothing",
                "now", "nowhere", "of", "off", "often", "on", "once",
                "one", "only", "onto", "or", "other", "others",
                "otherwise", "our", "ours", "ourselves", "out", "over",
                "own", "part", "per", "perhaps", "please", "put",
                "rather", "re", "same", "see", "seem", "seemed",
                "seeming", "seems", "serious", "several", "she",
                "should", "show", "side", "since", "sincere", "six",
                "sixty", "so", "some", "somehow", "someone", "something",
                "sometime", "sometimes", "somewhere", "still", "such",
                "system", "take", "ten", "than", "that", "the", "their",
                "them", "themselves", "then", "thence", "there",
                "thereafter", "thereby", "therefore", "therein",
                "thereupon", "these", "they", "thick", "thin", "third",
                "this", "those", "though", "three", "through",
                "throughout", "thru", "thus", "to", "together", "too",
                "top", "toward", "towards", "twelve", "twenty", "two",
                "un", "under", "until", "up", "upon", "us", "very",
                "via", "was", "we", "well", "were", "what", "whatever",
                "when", "whence", "whenever", "where", "whereafter",
                "whereas", "whereby", "wherein", "whereupon", "wherever",
                "whether", "which", "while", "whither", "who", "whoever",
                "whole", "whom", "whose", "why", "will", "with",
                "within", "without", "would", "yet", "you", "your",
                "yours", "yourself", "yourselves" };
        Arrays.sort(words);
        stopWords.put("en", words);

        // german (de)
        words = new String[] { "ab", "bei", "da", "deshalb", "ein",
                "für", "haben", "hier", "ich", "ja", "kann", "machen",
                "muesste", "nach", "oder", "seid", "sonst", "und", "vom",
                "wann", "wenn", "wie", "zu", "bin", "eines", "hat",
                "manche", "solches", "an", "anderm", "bis", "das",
                "deinem", "demselben", "dir", "doch", "einig", "er",
                "eurer", "hatte", "ihnen", "ihre", "ins", "jenen",
                "keinen", "manchem", "meinen", "nichts", "seine", "soll",
                "unserm", "welche", "werden", "wollte", "während",
                "alle", "allem", "allen", "aller", "alles", "als",
                "also", "am", "ander", "andere", "anderem", "anderen",
                "anderer", "anderes", "andern", "anderr", "anders",
                "auch", "auf", "aus", "bist", "bsp.", "daher", "damit",
                "dann", "dasselbe", "dazu", "daß", "dein", "deine",
                "deinen", "deiner", "deines", "dem", "den", "denn",
                "denselben", "der", "derer", "derselbe", "derselben",
                "des", "desselben", "dessen", "dich", "die", "dies",
                "diese", "dieselbe", "dieselben", "diesem", "diesen",
                "dieser", "dieses", "dort", "du", "durch", "eine",
                "einem", "einen", "einer", "einige", "einigem",
                "einigen", "einiger", "einiges", "einmal", "es", "etwas",
                "euch", "euer", "eure", "eurem", "euren", "eures",
                "ganz", "ganze", "ganzen", "ganzer", "ganzes", "gegen",
                "gemacht", "gesagt", "gesehen", "gewesen", "gewollt",
                "hab", "habe", "hatten", "hin", "hinter", "ihm", "ihn",
                "ihr", "ihrem", "ihren", "ihrer", "ihres", "im", "in",
                "indem", "ist", "jede", "jedem", "jeden", "jeder",
                "jedes", "jene", "jenem", "jener", "jenes", "jetzt",
                "kein", "keine", "keinem", "keiner", "keines", "konnte",
                "können", "könnte", "mache", "machst", "macht", "machte",
                "machten", "man", "manchen", "mancher", "manches",
                "mein", "meine", "meinem", "meiner", "meines", "mich",
                "mir", "mit", "muss", "musste", "müßt", "nicht", "noch",
                "nun", "nur", "ob", "ohne", "sage", "sagen", "sagt",
                "sagte", "sagten", "sagtest", "sehe", "sehen", "sehr",
                "seht", "sein", "seinem", "seinen", "seiner", "seines",
                "selbst", "sich", "sicher", "sie", "sind", "so",
                "solche", "solchem", "solchen", "solcher", "sollte",
                "sondern", "um", "uns", "unse", "unsen", "unser",
                "unses", "unter", "viel", "von", "vor", "war", "waren",
                "warst", "was", "weg", "weil", "weiter", "welchem",
                "welchen", "welcher", "welches", "werde", "wieder",
                "will", "wir", "wird", "wirst", "wo", "wolle", "wollen",
                "wollt", "wollten", "wolltest", "wolltet", "würde",
                "würden", "z.B.", "zum", "zur", "zwar", "zwischen",
                "über", "aber", "abgerufen", "abgerufene", "abgerufener",
                "abgerufenes", "acht", "acute", "allein", "allerdings",
                "allerlei", "allg", "allgemein", "allmählich", "allzu",
                "alsbald", "amp", "and", "andererseits", "andernfalls",
                "anerkannt", "anerkannte", "anerkannter", "anerkanntes",
                "anfangen", "anfing", "angefangen", "angesetze",
                "angesetzt", "angesetzten", "angesetzter", "ansetzen",
                "anstatt", "arbeiten", "aufgehört", "aufgrund",
                "aufhören", "aufhörte", "aufzusuchen", "ausdrücken",
                "ausdrückt", "ausdrückte", "ausgenommen", "ausser",
                "ausserdem", "author", "autor", "außen", "außer",
                "außerdem", "außerhalb", "background", "bald",
                "bearbeite", "bearbeiten", "bearbeitete", "bearbeiteten",
                "bedarf", "bedurfte", "bedürfen", "been", "befragen",
                "befragte", "befragten", "befragter", "begann",
                "beginnen", "begonnen", "behalten", "behielt", "beide",
                "beiden", "beiderlei", "beides", "beim", "beinahe",
                "beitragen", "beitrugen", "bekannt", "bekannte",
                "bekannter", "bekennen", "benutzt", "bereits",
                "berichten", "berichtet", "berichtete", "berichteten",
                "besonders", "besser", "bestehen", "besteht",
                "beträchtlich", "bevor", "bezüglich", "bietet", "bisher",
                "bislang", "biz", "bleiben", "blieb", "bloss", "bloß",
                "border", "brachte", "brachten", "brauchen", "braucht",
                "bringen", "bräuchte", "bzw", "böden", "ca", "ca.",
                "collapsed", "com", "comment", "content", "da?", "dabei",
                "dadurch", "dafür", "dagegen", "dahin", "damals",
                "danach", "daneben", "dank", "danke", "danken", "dannen",
                "daran", "darauf", "daraus", "darf", "darfst", "darin",
                "darum", "darunter", "darüber", "darüberhinaus", "dass",
                "davon", "davor", "demnach", "denen", "dennoch",
                "derart", "derartig", "derem", "deren", "derjenige",
                "derjenigen", "derzeit", "desto", "deswegen",
                "diejenige", "diesseits", "dinge", "direkt", "direkte",
                "direkten", "direkter", "doc", "doppelt", "dorther",
                "dorthin", "drauf", "drei", "dreißig", "drin", "dritte",
                "drunter", "drüber", "dunklen", "durchaus", "durfte",
                "durften", "dürfen", "dürfte", "eben", "ebenfalls",
                "ebenso", "ehe", "eher", "eigenen", "eigenes",
                "eigentlich", "einbaün", "einerseits", "einfach",
                "einführen", "einführte", "einführten", "eingesetzt",
                "einigermaßen", "eins", "einseitig", "einseitige",
                "einseitigen", "einseitiger", "einst", "einstmals",
                "einzig", "elf", "ende", "entsprechend", "entweder",
                "ergänze", "ergänzen", "ergänzte", "ergänzten",
                "erhalten", "erhielt", "erhielten", "erhält", "erneut",
                "erst", "erste", "ersten", "erster", "eröffne",
                "eröffnen", "eröffnet", "eröffnete", "eröffnetes", "etc",
                "etliche", "etwa", "fall", "falls", "fand", "fast",
                "ferner", "finden", "findest", "findet", "folgende",
                "folgenden", "folgender", "folgendes", "folglich", "for",
                "fordern", "fordert", "forderte", "forderten",
                "fortsetzen", "fortsetzt", "fortsetzte", "fortsetzten",
                "fragte", "frau", "frei", "freie", "freier", "freies",
                "fuer", "fünf", "gab", "ganzem", "gar", "gbr", "geb",
                "geben", "geblieben", "gebracht", "gedurft", "geehrt",
                "geehrte", "geehrten", "geehrter", "gefallen", "gefiel",
                "gefälligst", "gefällt", "gegeben", "gehabt", "gehen",
                "geht", "gekommen", "gekonnt", "gemocht", "gemäss",
                "genommen", "genug", "gern", "gestern", "gestrige",
                "getan", "geteilt", "geteilte", "getragen",
                "gewissermaßen", "geworden", "ggf", "gib", "gibt",
                "gleich", "gleichwohl", "gleichzeitig",
                "glücklicherweise", "gmbh", "gratulieren", "gratuliert",
                "gratulierte", "gute", "guten", "gängig", "gängige",
                "gängigen", "gängiger", "gängiges", "gänzlich", "haette",
                "halb", "hallo", "hast", "hattest", "hattet", "heraus",
                "herein", "heute", "heutige", "hiermit", "hiesige",
                "hinein", "hinten", "hinterher", "hoch", "html", "http",
                "hundert", "hätt", "hätte", "hätten", "höchstens",
                "igitt", "image", "immer", "immerhin", "important",
                "indessen", "info", "infolge", "innen", "innerhalb",
                "insofern", "inzwischen", "irgend", "irgendeine",
                "irgendwas", "irgendwen", "irgendwer", "irgendwie",
                "irgendwo", "je", "jed", "jedenfalls", "jederlei",
                "jedoch", "jemand", "jenseits", "jährig", "jährige",
                "jährigen", "jähriges", "kam", "kannst", "kaum",
                "kei nes", "keinerlei", "keineswegs", "klar", "klare",
                "klaren", "klares", "klein", "kleinen", "kleiner",
                "kleines", "koennen", "koennt", "koennte", "koennten",
                "komme", "kommen", "kommt", "konkret", "konkrete",
                "konkreten", "konkreter", "konkretes", "konnten", "könn",
                "könnt", "könnten", "künftig", "lag", "lagen", "langsam",
                "lassen", "laut", "lediglich", "leer", "legen", "legte",
                "legten", "leicht", "leider", "lesen", "letze",
                "letzten", "letztendlich", "letztens", "letztes",
                "letztlich", "lichten", "liegt", "liest", "links",
                "längst", "längstens", "mag", "magst", "mal",
                "mancherorts", "manchmal", "mann", "margin", "med",
                "mehr", "mehrere", "meist", "meiste", "meisten", "meta",
                "mindestens", "mithin", "mochte", "morgen", "morgige",
                "muessen", "muesst", "musst", "mussten", "muß", "mußt",
                "möchte", "möchten", "möchtest", "mögen", "möglich",
                "mögliche", "möglichen", "möglicher", "möglicherweise",
                "müssen", "müsste", "müssten", "müßte", "nachdem",
                "nacher", "nachhinein", "nahm", "natürlich", "ncht",
                "neben", "nebenan", "nehmen", "nein", "neu", "neue",
                "neuem", "neuen", "neuer", "neues", "neun", "nie",
                "niemals", "niemand", "nimm", "nimmer", "nimmt",
                "nirgends", "nirgendwo", "nter", "nutzen", "nutzt",
                "nutzung", "nächste", "nämlich", "nötigenfalls", "nützt",
                "oben", "oberhalb", "obgleich", "obschon", "obwohl",
                "oft", "online", "org", "padding", "per", "pfui",
                "plötzlich", "pro", "reagiere", "reagieren", "reagiert",
                "reagierte", "rechts", "regelmäßig", "rief", "rund",
                "sang", "sangen", "schlechter", "schließlich", "schnell",
                "schon", "schreibe", "schreiben", "schreibens",
                "schreiber", "schwierig", "schätzen", "schätzt",
                "schätzte", "schätzten", "sechs", "sect", "sehrwohl",
                "sei", "seit", "seitdem", "seite", "seiten", "seither",
                "selber", "senke", "senken", "senkt", "senkte",
                "senkten", "setzen", "setzt", "setzte", "setzten",
                "sicherlich", "sieben", "siebte", "siehe", "sieht",
                "singen", "singt", "sobald", "sodaß", "soeben", "sofern",
                "sofort", "sog", "sogar", "solange", "solc hen", "solch",
                "sollen", "sollst", "sollt", "sollten", "solltest",
                "somit", "sonstwo", "sooft", "soviel", "soweit", "sowie",
                "sowohl", "spielen", "später", "startet", "startete",
                "starteten", "statt", "stattdessen", "steht", "steige",
                "steigen", "steigt", "stets", "stieg", "stiegen", "such",
                "suchen", "sämtliche", "tages", "tat", "tatsächlich",
                "tatsächlichen", "tatsächlicher", "tatsächliches",
                "tausend", "teile", "teilen", "teilte", "teilten",
                "titel", "total", "trage", "tragen", "trotzdem", "trug",
                "trägt", "tun", "tust", "tut", "txt", "tät", "ueber",
                "umso", "unbedingt", "ungefähr", "unmöglich",
                "unmögliche", "unmöglichen", "unmöglicher", "unnötig",
                "unsem", "unser", "unsere", "unserem", "unseren",
                "unserer", "unseres", "unten", "unterbrach",
                "unterbrechen", "unterhalb", "unwichtig", "usw", "var",
                "vergangen", "vergangene", "vergangener", "vergangenes",
                "vermag", "vermutlich", "vermögen", "verrate",
                "verraten", "verriet", "verrieten", "version",
                "versorge", "versorgen", "versorgt", "versorgte",
                "versorgten", "versorgtes", "veröffentlichen",
                "veröffentlicher", "veröffentlicht", "veröffentlichte",
                "veröffentlichten", "veröffentlichtes", "viele",
                "vielen", "vieler", "vieles", "vielleicht", "vielmals",
                "vier", "vollständig", "voran", "vorbei", "vorgestern",
                "vorher", "vorne", "vorüber", "völlig", "während",
                "wachen", "waere", "warum", "weder", "wegen", "weitere",
                "weiterem", "weiteren", "weiterer", "weiteres",
                "weiterhin", "weiß", "wem", "wen", "wenig", "wenige",
                "weniger", "wenigstens", "wenngleich", "wer", "werdet",
                "weshalb", "wessen", "wichtig", "wieso", "wieviel",
                "wiewohl", "willst", "wirklich", "wodurch", "wogegen",
                "woher", "wohin", "wohingegen", "wohl", "wohlweislich",
                "womit", "woraufhin", "woraus", "worin", "wurde",
                "wurden", "währenddessen", "wär", "wäre", "wären",
                "zahlreich", "zehn", "zeitweise", "ziehen", "zieht",
                "zog", "zogen", "zudem", "zuerst", "zufolge", "zugleich",
                "zuletzt", "zumal", "zurück", "zusammen", "zuviel",
                "zwanzig", "zwei", "zwölf", "ähnlich", "übel", "überall",
                "überallhin", "überdies", "übermorgen", "übrig",
                "übrigens" };
        Arrays.sort(words);
        stopWords.put("de", words);

        // spanish (es)
        words = new String[] { "era", "erais", "eran", "eras", "eres",
                "es", "está", "estaba", "estabais", "estaban", "estabas",
                "estad", "estada", "estadas", "estado", "estados",
                "estamos", "estando", "estaré", "estaría", "estaríais",
                "estaríamos", "estarían", "estarías", "estaremos",
                "estaréis", "estarán", "estarás", "estábamos", "estemos",
                "estáis", "están", "estoy", "estás", "estuve",
                "estuviera", "estuvierais", "estuvieran", "estuvieras",
                "estuvieron", "estuviese", "estuvieseis", "estuviesen",
                "estuvieses", "estuvimos", "estuviéramos",
                "estuviésemos", "estuviste", "estuvisteis", "estuvo",
                "fue", "fuera", "fuerais", "fueran", "fueras", "fueron",
                "fuese", "fueseis", "fuesen", "fueses", "fui", "fuimos",
                "fuiste", "fuisteis", "fuéramos", "fuésemos", "ha",
                "había", "habíais", "habíamos", "habían", "habías",
                "habida", "habidas", "habido", "habidos", "habiendo",
                "habéis", "habré", "habría", "habríais", "habríamos",
                "habrían", "habrías", "habremos", "habréis", "habrán",
                "habrás", "han", "has", "haya", "hayamos", "hayan",
                "hayas", "hayáis", "he", "hemos", "hube", "hubiera",
                "hubierais", "hubieran", "hubieras", "hubieron",
                "hubiese", "hubieseis", "hubiesen", "hubieses",
                "hubimos", "hubiéramos", "hubiésemos", "hubiste",
                "hubisteis", "hubo", "éramos", "sea", "seamos", "sean",
                "seas", "seáis", "seré", "sería", "seríais", "seríamos",
                "serían", "serías", "seremos", "seréis", "serán",
                "serás", "sido", "siendo", "sois", "somos", "son", "soy",
                "tenía", "teníais", "teníamos", "tenían", "tenías",
                "tendré", "tendría", "tendríais", "tendríamos",
                "tendrían", "tendrías", "tendremos", "tendréis",
                "tendrán", "tendrás", "tened", "tenemos", "tenga",
                "tengamos", "tengan", "tengas", "tengáis", "tengo",
                "tenida", "tenidas", "tenido", "tenidos", "teniendo",
                "tenéis", "tiene", "tienen", "tienes", "tuve", "tuviera",
                "tuvierais", "tuvieran", "tuvieras", "tuvieron",
                "tuviese", "tuvieseis", "tuviesen", "tuvieses",
                "tuvimos", "tuviéramos", "tuviésemos", "tuviste",
                "tuvisteis", "tuvo" };
        Arrays.sort(words);
        stopWords.put("es", words);

        // french (fr)
        words = new String[] { "à", "ai", "aie", "aient", "aies", "ait",
                "as", "au", "aura", "aurai", "auraient", "aurais",
                "aurait", "auras", "aurez", "auriez", "aurions",
                "aurons", "auront", "aux", "avaient", "avais", "avait",
                "avec", "avez", "aviez", "avions", "avons", "ayant",
                "ayante", "ayantes", "ayants", "ayez", "ayons", "c",
                "ce", "ces", "d", "dans", "de", "des", "du", "elle",
                "eûmes", "en", "es", "est", "et", "eût", "eûtes", "eu",
                "eue", "eues", "eurent", "eus", "eusse", "eussent",
                "eusses", "eussiez", "eussions", "eut", "eux", "fûmes",
                "fût", "fûtes", "furent", "fus", "fusse", "fussent",
                "fusses", "fussiez", "fussions", "fut", "il", "j", "je",
                "l", "la", "le", "leur", "lui", "m", "ma", "mais", "me",
                "mes", "même", "moi", "mon", "n", "ne", "nos", "notre",
                "nous", "on", "ont", "ou", "par", "pas", "pour", "qu",
                "que", "qui", "s", "sa", "se", "sera", "serai",
                "seraient", "serais", "serait", "seras", "serez",
                "seriez", "serions", "serons", "seront", "ses", "soient",
                "sois", "soit", "sommes", "son", "sont", "soyez",
                "soyons", "suis", "sur", "t", "été", "ta", "étaient",
                "étais", "était", "étant", "étante", "étantes", "étants",
                "te", "étée", "tes", "êtes", "étées", "étiez", "étions",
                "toi", "ton", "étés", "tu", "un", "une", "vos", "votre",
                "vous", "y" };
        Arrays.sort(words);
        stopWords.put("fr", words);

        // italian (it)
        words = new String[] { "è", "abbia", "abbiamo", "abbiano",
                "abbiate", "avemmo", "avendo", "avesse", "avessero",
                "avessi", "avessimo", "aveste", "avesti", "avete",
                "aveva", "avevamo", "avevano", "avevate", "avevi",
                "avevo", "avrò", "avrai", "avranno", "avrebbe",
                "avrebbero", "avrei", "avremmo", "avremo", "avreste",
                "avresti", "avrete", "avuta", "avute", "avuti", "avuto",
                "ebbe", "ebbero", "ebbi", "era", "erano", "eravamo",
                "eravate", "eri", "ero", "essendo", "faccia", "facciamo",
                "facciano", "facciate", "faccio", "facemmo", "facendo",
                "facesse", "facessero", "facessi", "facessimo",
                "faceste", "facesti", "faceva", "facevamo", "facevano",
                "facevate", "facevi", "facevo", "fai", "fanno", "farò",
                "farai", "faranno", "farebbe", "farebbero", "farei",
                "faremmo", "faremo", "fareste", "faresti", "farete",
                "fece", "fecero", "feci", "fosse", "fossero", "fossi",
                "fossimo", "foste", "fosti", "fu", "fui", "fummo",
                "furono", "ha", "hai", "hanno", "ho", "sarò", "sarai",
                "saranno", "sarebbe", "sarebbero", "sarei", "saremmo",
                "saremo", "sareste", "saresti", "sarete", "sei", "sia",
                "siamo", "siano", "siate", "siete", "sono", "sta",
                "stai", "stando", "stanno", "starò", "starai",
                "staranno", "starebbe", "starebbero", "starei",
                "staremmo", "staremo", "stareste", "staresti", "starete",
                "stava", "stavamo", "stavano", "stavate", "stavi",
                "stavo", "stemmo", "stesse", "stessero", "stessi",
                "stessimo", "steste", "stesti", "stette", "stettero",
                "stetti", "stia", "stiamo", "stiano", "stiate", "sto" };
        Arrays.sort(words);
        stopWords.put("it", words);

        // dutch (nl)
        words = new String[] { "aan", "al", "alles", "als", "altijd",
                "andere", "ben", "bij", "daar", "dan", "dat", "de",
                "der", "deze", "die", "dit", "doch", "doen", "door",
                "dus", "een", "eens", "en", "er", "ge", "geen",
                "geweest", "haar", "had", "heb", "hebben", "heeft",
                "hem", "het", "hier", "hij", "hoe", "hun", "iemand",
                "iets", "ik", "in", "is", "ja", "je", "kan", "kon",
                "kunnen", "maar", "me", "meer", "men", "met", "mij",
                "mijn", "moet", "na", "naar", "niet", "niets", "nog",
                "nu", "of", "om", "omdat", "onder", "ons", "ook", "op",
                "over", "reeds", "te", "tegen", "toch", "toen", "tot",
                "u", "uit", "uw", "van", "veel", "voor", "want", "waren",
                "was", "wat", "werd", "wezen", "wie", "wil", "worden",
                "wordt", "zal", "ze", "zelf", "zich", "zij", "zijn",
                "zo", "zonder", "zou" };
        Arrays.sort(words);
        stopWords.put("nl", words);

        // TODO: language-specific stop words for more languages,
        // especially when we have support for non-roman terms

        // set default stop words
        languageStopWords = stopWords.get(DEFAULT_LANGUAGE);
    }

    /**
     * @param _parent
     * @param _paths
     * @return
     */
    protected List<File> expandPaths(File _parent, String[] _paths) {
        List<File> paths = new Vector<File>(_paths.length);
        File f;
        for (int i = 0; i < _paths.length; i++) {
            f = (null == _parent) ? new File(_paths[i]) : new File(
                    _parent, _paths[i]);
            if (!f.canRead()) {
                logger.warning("skipping " + _paths[i] + ": cannot read");
                continue;
            }
            if (f.isFile()) {
                paths.add(f);
                continue;
            }
            if (f.isDirectory()) {
                paths.addAll(expandPaths(f, f.list()));
                continue;
            }
            logger.warning("skipping " + _paths[i] + ": not a file");
        }
        return paths;
    }

    /**
     * @param args
     */
    protected void initialize(String[] args) {
        Arrays.sort(nameList);

        paths = expandPaths(null, args);
        pathIterator = paths.iterator();
        if (!pathIterator.hasNext()) {
            throw new FatalException("no input paths found");
        }

        try {
            xpp = config.getXppFactory().newPullParser();
            // TODO feature isn't supported by xpp3 - look at xpp5?
            // xpp.setFeature(XmlPullParser.FEATURE_DETECT_ENCODING, true);
            // TODO feature isn't supported by xpp3 - look at xpp5?
            // xpp.setFeature(XmlPullParser.FEATURE_PROCESS_DOCDECL, true);
            xpp
                    .setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES,
                            true);
        } catch (XmlPullParserException e) {
            throw new FatalException(e);
        }
    }

    /**
     * @param _random
     * @param _array
     * 
     *            cf http://en.wikipedia.org/wiki/Knuth_shuffle
     */
    public static void shuffle(Random _random, Object[] _array) {
        // items left to shuffle (loop invariant)
        Object tmp;
        int n = _array.length;
        while (n > 1) {
            n--;
            // n is now the last pertinent index
            // 0 <= k <= n
            int k = _random.nextInt(n + 1);
            // swap
            tmp = _array[k];
            _array[k] = _array[n];
            _array[n] = tmp;
        }
    }

    /**
     * @param _seed
     */
    public void shuffle(int _seed) {
        logger.info("seed " + _seed);
        // repeatable seed value
        Random random = new Random(_seed);
        synchronized (readerMutex) {
            reader = null;
            // shuffle the paths - three times for a good shuffle
            File[] pathArray = paths.toArray(new File[0]);
            logger.info(String.format("pathArray[0] = %s", pathArray[0]));
            for (int i = 0; i < 3; i++) {
                shuffle(random, pathArray);
                logger.info(String.format("pathArray[0] = %s",
                        pathArray[0]));
            }
            paths = Arrays.asList(pathArray);
            pathIterator = paths.iterator();
            logger.info(String.format("paths[0] = %s", paths.get(0)));
        }
    }

    protected void prepareReader() {
        if (null == reader) {
            synchronized (readerMutex) {
                try {
                    while (null == reader) {
                        if (null != zipInputStream) {
                            // old entry has been closed already
                            ZipEntry entry = zipInputStream
                                    .getNextEntry();
                            if (null == entry) {
                                zipInputStream = null;
                                continue;
                            }
                            if (entry.isDirectory()) {
                                continue;
                            }
                            logger.fine("preparing reader from entry: "
                                    + entry.getName());
                            reader = new InputStreamReader(zipInputStream);
                            continue;
                        }

                        // no reader and no zip? next path....
                        if (!pathIterator.hasNext()) {
                            // reset to beginning
                            pathIterator = paths.iterator();
                        }
                        CharsetDecoder decoder = Charset.forName(
                                Configuration.INPUT_ENCODING_DEFAULT)
                                .newDecoder();
                        File path = pathIterator.next();
                        String canonicalPath = path.getCanonicalPath();
                        logger.info("preparing reader: " + canonicalPath);
                        if (canonicalPath.endsWith(".zip")) {
                            zipInputStream = new ZipInputStream(
                                    new FileInputStream(path));
                            continue;
                        }
                        reader = new InputStreamReader(
                                new FileInputStream(path), decoder);
                    }
                    xpp.setInput(reader);
                } catch (IOException e) {
                    throw new FatalException(e);
                } catch (XmlPullParserException e) {
                    throw new FatalException(e);
                }
            }
        }
    }

    /**
     * @throws IOException
     */
    protected void resetReader() throws IOException {
        if (null != reader) {
            synchronized (readerMutex) {
                if (null != reader) {
                    if (null == zipInputStream) {
                        // will this leak?
                        // we have to close the underlying zip stream
                        reader.close();
                    }
                    reader = null;
                    elementStack.clear();
                }
            }
        }
        prepareReader();
    }

    /**
     * @param _config
     * @param _paths
     */
    public AbstractTermGenerator(Configuration _config, String[] _paths) {
        config = _config;
        logger = config.getLogger();

        initialize(_paths);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.TermGenerator#nextWord()
     */
    public String nextWord() {
        return nextWords(1)[0];
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.TermGenerator#nextWords(int)
     */
    public String[] nextWords(int size) {
        String[] words;
        int eventType;
        long count = 0;
        synchronized (readerMutex) {
            try {
                while (size > buffer.size()) {
                    if (null == reader) {
                        prepareReader();
                    }
                    try {
                        eventType = xpp.nextToken();
                    } catch (XmlPullParserException e) {
                        // move on to the next available entry
                        logger.logException(e);
                        resetReader();
                        continue;
                    }
                    count++;
                    if (count > LIMIT) {
                        throw new FatalException("infinite loop?");
                    }
                    String name = xpp.getName();
                    switch (eventType) {
                    case XmlPullParser.END_DOCUMENT:
                        // logger.fine("end of document");
                        resetReader();
                        break;
                    case XmlPullParser.END_TAG:
                        if (0 == elementStack.size()) {
                            throw new FatalException(
                                    "end tag on empty stack! "
                                            + xpp
                                                    .getPositionDescription());
                        }
                        String pop = elementStack.pop();
                        if (name != pop) {
                            throw new FatalException(
                                    "element name mismatch, got "
                                            + name
                                            + " expected "
                                            + pop
                                            + xpp
                                                    .getPositionDescription());
                        }
                        break;
                    case XmlPullParser.START_TAG:
                        elementStack.push(name);
                        // logger.finer("element " + element);
                        checkLanguage();
                        break;
                    case XmlPullParser.TEXT:
                        if (matches()) {
                            // buffer the word tokens
                            String rawText = xpp.getText().trim()
                                    .toLowerCase();
                            if (rawText.length() > 0) {
                                logger.finest("text = " + rawText);
                                String[] text = rawText
                                        .split(SPLIT_PATTERN);
                                for (int i = 0; i < text.length; i++) {
                                    // terms must be at least 5 characters
                                    if (5 > text[i].length()) {
                                        continue;
                                    }
                                    // terms must not contain numbers
                                    if (!text[i].matches("^\\D+$")) {
                                        continue;
                                    }
                                    // terms must not be language stop words
                                    if (null != languageStopWords
                                            && -1 < Arrays.binarySearch(
                                                    languageStopWords,
                                                    text[i])) {
                                        continue;
                                    }
                                    // terms must not be repetitive
                                    if (repetitionQueue.contains(text[i])) {
                                        continue;
                                    }
                                    buffer.add(text[i]);
                                    repetitionQueue.add(text[i]);
                                    if (repetitionQueue.size() > REPETITION_QUEUE_SIZE) {
                                        repetitionQueue.removeFirst();
                                    }
                                }
                                // logger.finer("buffer size = "
                                // + buffer.size());
                            }
                        }
                        break;
                    default:
                        break;
                    }
                }
            } catch (IOException e) {
                throw new FatalException(e);
            }
            words = new String[size];
            for (int i = 0; i < size; i++) {
                words[i] = buffer.remove(0);
            }
        }
        return words;
    }

    /**
     * 
     */
    private void checkLanguage() {
        // use: caller asks for word(s), then checks language of those words
        // requirement: all words in the buffer should match the language field
        // implementation: if there is an xml:lang attribute,
        // and it is not the same as the language field,
        // then clear the buffer and update the language field.
        String thisLanguage = xpp.getAttributeValue(XML_LANG_NAMESPACE,
                XML_LANG_NAME);
        if (null == thisLanguage) {
            return;
        }
        if (language.equals(thisLanguage)) {
            return;
        }
        logger.fine(String.format("language changed from %s to %s",
                language, thisLanguage));
        language = thisLanguage;
        languageStopWords = stopWords.get(language);
        // languages with no stop words configured will use English
        // that should be ok, since they will mostly be non-roman
        // and so any roman words will probably be English anyway.
        if (null == languageStopWords) {
            languageStopWords = stopWords.get(DEFAULT_LANGUAGE);
        }
        buffer.clear();
        repetitionQueue.clear();
    }

    /**
     * @return
     */
    private boolean matches() {
        if (null == nameList) {
            throw new NullPointerException("null nameList at "
                    + xpp.getPositionDescription());
        }
        if (0 == nameList.length) {
            throw new NullPointerException("empty nameList at "
                    + xpp.getPositionDescription());
        }
        if (null == elementStack) {
            throw new NullPointerException("null element at "
                    + xpp.getPositionDescription());
        }
        String element;
        Iterator<String> iter = elementStack.iterator();
        while (iter.hasNext()) {
            element = iter.next();
            if (Arrays.binarySearch(nameList, element) >= 0) {
                return true;
            }
        }
        return false;
    }

    public String getLanguage() {
        return language;
    }

}
