/**
 * Copyright (c) 2009-2010 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.lang.reflect.Field;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 */
public class Configuration {

    public static final String DEFAULT_SUFFIX = "_DEFAULT";

    public static final String KEY_SUFFIX = "_KEY";

    protected static SimpleLogger logger = null;

    protected Map<String, Object> defaults = new HashMap<String, Object>();

    protected Properties properties = new Properties();

    protected XmlPullParserFactory factory;

    protected Date start;

    public static final int SLEEP_TIME = 500;

    // external property keys and default values

    public static final String INPUT_ENCODING_DEFAULT = "UTF-8";

    public static final String INPUT_ENCODING_KEY = "INPUT_ENCODING";

    public static final String LOG_LEVEL_DEFAULT = "INFO";

    public static final String LOG_LEVEL_KEY = "LOG_LEVEL";

    public static final String TEST_TIME_KEY = "TEST_TIME";

    public static final String TEST_TIME_DEFAULT = "" + 3600;

    public static final String WORKERS_PER_CONTROLLER_KEY = "WORKERS_PER_CONTROLLER";

    public static final String WORKERS_PER_CONTROLLER_DEFAULT = "" + 1;

    public static final String GENERATOR_PATH_KEY = "GENERATOR_PATH";

    public static final String GENERATOR_PATH_DEFAULT = "wiki-zip";

    public static final String CONNECTION_STRING_KEY = "CONNECTION_STRING";

    public static final String CONNECTION_STRING_DEFAULT = "http://localhost:8000/";

    public static final String FATAL_ERRORS_KEY = "FATAL_ERRORS";

    public static final String FATAL_ERRORS_DEFAULT = "false";

    public static final String WORKER_CLASS_KEY = "WORKER_CLASS";

    public static final String WORKER_CLASS_DEFAULT = HttpWorker.class
            .getCanonicalName();

    private String[] connections;

    private int connectionsLength;

    private int connectionsCount;

    /**
     * @param _config
     */
    public Configuration(Configuration _config) {
        load(_config.getProperties());
    }

    /**
     * 
     */
    public Configuration() {
        // default constructor
    }

    protected void setDefaults() throws IllegalArgumentException,
            SecurityException, IllegalAccessException,
            NoSuchFieldException {
        Field[] fields = this.getClass().getFields();
        String name, key;
        for (int i = 0; i < fields.length; i++) {
            name = fields[i].getName();
            if (name.endsWith(KEY_SUFFIX)) {
                key = (String) fields[i].get(this);
                if (!defaults.containsKey(key)) {
                    defaults.put(key, null);
                    logger.fine(key + "=(null)");
                }
                continue;
            }
            if (name.endsWith(DEFAULT_SUFFIX)) {
                // the true key is the value of the key-named field
                key = (String) this.getClass().getField(
                        name.substring(0, name.length()
                                - DEFAULT_SUFFIX.length())
                                + KEY_SUFFIX).get(this);
                defaults.put(key, fields[i].get(this));
                logger.fine(key + "=" + defaults.get(key));
                continue;
            }
        }
    }

    /**
     * @param _props
     */
    public void load(Properties _props) {
        properties.putAll(_props);
    }

    /**
     * @param _stream
     * @throws IOException
     */
    public void load(InputStream _stream) throws IOException {
        Properties newProperties = new Properties();
        newProperties.load(_stream);
        load(newProperties);
    }

    /**
     * validate user-defined property names and apply defaults
     */
    protected void validateProperties() {
        Properties validated = new Properties();
        Enumeration<?> keys = properties.propertyNames();
        // ignore known patterns from System properties
        String ignorePatterns = "^(file|java|line|os|path|sun|user)\\..+";
        String key, value;
        while (keys.hasMoreElements()) {
            key = (String) keys.nextElement();
            // known jre pattern
            if (key.matches(ignorePatterns)) {
                logger.fine("known system key: ignoring " + key);
                continue;
            }
            // key is unknown
            if (!defaults.containsKey(key)) {
                logger.warning("unknown key: skipping " + key);
                continue;
            }
            // known key
            value = properties.getProperty(key);
            logger.info("using " + key + "=" + value);
            validated.setProperty(key, value);
        }

        applyDefaults(validated);

        // use the validated properties
        properties = validated;
    }

    /**
     * @param _props
     */
    private void applyDefaults(Properties _props) {
        // apply default values to properties
        Iterator<String> iter = defaults.keySet().iterator();
        String key;
        Object value;
        while (iter.hasNext()) {
            key = iter.next();
            if (_props.containsKey(key)) {
                continue;
            }
            value = defaults.get(key);
            if (null != value) {
                logger.fine("applying default " + key + "=" + value);
                _props.setProperty(key.toString(), value.toString());
            }
        }
    }

    public void configure() {
        // set up the logger early, for verbose configuration output
        if (null == logger) {
            logger = SimpleLogger.getSimpleLogger();
        }
        logger.configureLogger(properties);

        // time at which this configuration object was finalized
        start = new Date();

        try {
            setDefaults();
        } catch (Exception e) {
            // crude, but this is a configuration-time error
            throw new FatalException(e);
        }
        validateProperties();
        
        String connList = properties.getProperty(CONNECTION_STRING_KEY);
        connections = connList.split("\\s+");
        connectionsLength = connections.length;
        connectionsCount = 0;
    }

    /**
     * @return
     */
    public SimpleLogger getLogger() {
        return logger;
    }

    /**
     * @return
     */
    public Properties getProperties() {
        return properties;
    }

    /**
     * @param _logger
     */
    public void setLogger(SimpleLogger _logger) {
        logger = _logger;
    }

    /**
     * 
     * @throws XmlPullParserException
     * @return
     */
    public XmlPullParserFactory getXppFactory()
            throws XmlPullParserException {
        // get a new factory
        if (null == factory) {
            factory = XmlPullParserFactory.newInstance(properties
                    .getProperty(XmlPullParserFactory.PROPERTY_NAME),
                    null);
            factory.setNamespaceAware(true);
        }
        return factory;
    }

    /**
     * @return
     */
    public long getTestTimeNanos() {
        return Timer.NANOSECONDS_PER_SECOND
                * Long.valueOf(properties.getProperty(TEST_TIME_KEY));
    }

    /**
     * @return
     */
    public int getWorkersPerController() {
        return Integer.valueOf(properties
                .getProperty(WORKERS_PER_CONTROLLER_KEY));
    }

    /**
     * @return
     */
    public Date getStart() {
        return start;
    }

    /**
     * @return
     */
    public String getGeneratorPath() {
        return properties.getProperty(GENERATOR_PATH_KEY);
    }

    /**
     * @return
     */
    public String getConnectionString() {
        if (1 == connectionsLength) {
            return connections[0];
        }
        // round-robin for load balancing - atomic?
        return connections[connectionsCount++ % connectionsLength];
    }

    /**
     * @return
     */
    public boolean isFatalErrors() {
        return Boolean.valueOf(properties.getProperty(FATAL_ERRORS_KEY));
    }

    /**
     * @param out
     */
    public void list(PrintStream out) {
        properties.list(out);
    }

    /**
     * @return
     */
    public String getWorkerClassName() {
        return properties.getProperty(WORKER_CLASS_KEY);
    }

    /**
     * @return
     */
    public String getLogLevel() {
        return properties.getProperty(LOG_LEVEL_KEY);
    }

}
