/**
 * Copyright (c) 2009 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 *         Controller for worker threads, created and run by a Director.
 * 
 *         May be local to Director host
 * 
 *         TODO: May run in a remote jvm on another host
 * 
 */
public class Controller extends Thread {

    protected Configuration config;

    protected SimpleLogger logger;

    protected int index;

    protected boolean active = false;

    protected Director director;

    protected TermGenerator generator;

    protected String hostname;

    protected Timer timer;

    protected Worker[] workers;

    /**
     * @param _director
     * @param _index
     * @throws UnknownHostException
     */
    public Controller(Director _director, int _index)
            throws UnknownHostException {
        director = _director;
        index = _index;
        config = director.getConfig();
        logger = config.getLogger();
        logger.info("" + logger.getLevel());
        logger.info("" + config.getLogLevel());

        setName("Controller-" + index);
        logger.info(getName());

        hostname = InetAddress.getLocalHost().getHostName();
    }

    /**
     * @return
     */
    protected boolean pollWorkers() {
        boolean active = false;
        long requests;
        requests = 0;
        for (int i = 0; i < workers.length; i++) {
            if (workers[i].isAlive() && workers[i].isAlive()) {
                active = true;
            }
            requests += workers[i].getRequestCount();
        }
        timer.setEventCount(requests);
        return active;
    }

    @Override
    public void run() {
        active = true;
        workers = null;
        timer = new Timer();

        prepareGenerator();

        prepareWorkers();

        // monitor until all controllers have stopped
        while (active && pollWorkers()) {
            try {
                sleep(100);
            } catch (InterruptedException e) {
                // ignore
            }
        }
        timer.stop();
        active = false;
        director.interrupt();

        // final update
        pollWorkers();

        logger.info(getName() + " exiting");
    }

    /**
     * 
     */
    protected void prepareWorkers() {
        // prepare worker constructor
        String workerClassName = config.getWorkerClassName();
        logger.info("Worker class is " + workerClassName);
        try {
            Class<? extends Worker> workerClass = Class.forName(
                    workerClassName, true,
                    ClassLoader.getSystemClassLoader()).asSubclass(
                    Worker.class);
            Constructor<? extends Worker> workerConstructor = workerClass
                    .getConstructor(new Class[] { Controller.class,
                            int.class });

            // create and start worker threads
            int size = config.getWorkersPerController();
            workers = new Worker[size];
            Object[] args = new Object[2];
            args[0] = this;
            for (int i = 0; i < size; i++) {
                // use reflection-based constructor to build worker
                args[1] = i;
                workers[i] = workerConstructor.newInstance(args);
                workers[i].start();
            }
        } catch (ClassNotFoundException e) {
            throw new FatalException(workerClassName, e);
        } catch (IllegalArgumentException e) {
            throw new FatalException(workerClassName, e);
        } catch (InstantiationException e) {
            throw new FatalException(workerClassName, e);
        } catch (IllegalAccessException e) {
            throw new FatalException(workerClassName, e);
        } catch (InvocationTargetException e) {
            throw new FatalException(workerClassName, e);
        } catch (SecurityException e) {
            throw new FatalException(workerClassName, e);
        } catch (NoSuchMethodException e) {
            throw new FatalException(workerClassName, e);
        }
    }

    /**
     * 
     */
    protected void prepareGenerator() {
        String path = config.getGeneratorPath();
        File directory = new File(path);
        if (!directory.exists()) {
            FatalException e = new FatalException(
                    "cannot read path: " + path);
            director.reportException(e);
            throw (e);
        }
        generator = new WikipediaTermGenerator(config,
                new String[] { path });
        // prevent each controller from starting at the same terms
        // random seed based on controller index...
        generator.shuffle(getName().hashCode());
    }

    /**
     * @return
     */
    public boolean isActive() {
        return active;
    }

    /**
     * @return
     */
    public Configuration getConfig() {
        return config;
    }

    /**
     * @param e
     */
    public void reportException(Exception e) {
        if (e instanceof ActionException
                && ((ActionException) e).isTerse()) {
            logger.warning(e.getMessage());
        } else {
            logger.logException(e);
        }
        timer.incrementErrorCount();
        director.reportException(e);
    }

    /**
     * 
     */
    public void shutdown() {
        logger.warning("received shutdown signal");
        active = false;
        interrupt();
    }

    /**
     * @return
     */
    public TermGenerator getGenerator() {
        return generator;
    }

    /**
     * @return
     */
    public String getHostname() {
        return hostname;
    }

    /**
     * @return
     */
    public long getRequestCount() {
        return timer.getEventCount();
    }

    /**
     * @param index
     * @return
     * @throws IOException
     */
    public BufferedReader getResultsReader(int index) throws IOException {
        return new BufferedReader(workers[index].getResultsReader());
    }

    /**
     * @param j
     * @return
     */
    public String getWorkerName(int j) {
        return workers[j].getName();
    }

}
