/**
 * Copyright (c) 2009 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.BufferedInputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.Date;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 *         Logger for worker thread results
 * 
 */
public class ResultLogger implements Closeable {

    protected File log;

    protected byte[] readBuffer = new byte[32 * 1024];

    protected FileWriter writer;

    protected long index = 0;

    /**
     * @param _date
     * @param name
     * @throws IOException
     */
    public ResultLogger(Date _date, String name) throws IOException {
        // create new results file
        log = new File("ResultLogger/" + Utilities.formatDateTime(_date)
                + "/" + name);
        log.getParentFile().mkdirs();
        if (!log.createNewFile()) {
            throw new IOException("log file already exists: "
                    + log.getCanonicalPath());
        }
        writer = new FileWriter(log);
        writer.append("timestamp,index,description,"
                + "responseNanos,bytes,totalNanos\n");
    }

    /**
     * @param te
     * @param inputStream
     * @throws IOException
     */
    public void log(TimedEvent te, InputStream inputStream)
            throws IOException {
        // record bytes transferred
        te.increment(te.getDescription().length());
        // record time before handling results
        long response = te.getDuration();
        // handle results
        BufferedInputStream bin = null;
        try {
            bin = new BufferedInputStream(inputStream);
            int actual = 0;
            while ((actual = bin.read(readBuffer)) > -1) {
                te.increment(actual);
            }
        } finally {
            if (null != bin) {
                bin.close();
            }
        }

        // record time after handling results
        te.stop();

        // write to log file
        logMessage(te.getDescription(), response, te.getBytes(), te
                .getDuration());
    }

    /**
     * @param te
     * @param response
     * @throws IOException
     */
    public void log(TimedEvent te, long response) throws IOException {
        // record time after handling results
        te.stop();

        // write to log file
        logMessage(te.getDescription(), response, te.getBytes(), te
                .getDuration());
    }

    private void logMessage(String description, long response,
            long bytes, long duration) throws IOException {
        // CSV:
        // iso8601-dateTime,index,description,responseNanos,bytes,totalNanos
        writer.append(Utilities.formatDateTime() + "," + index + ","
                + "\"" + description.replaceAll("\"", "\\\"") + "\","
                + response + "," + bytes + "," + duration + "\n");
        index++;
    }

    /**
     * @throws IOException
     * 
     */
    public void close() throws IOException {
        writer.flush();
        writer.close();
        writer = null;
    }

    /**
     * @param te
     * @param e
     * @throws IOException
     */
    public void log(TimedEvent te, Exception e) throws IOException {
        // record time
        te.stop();

        // negative numbers are an error flag
        logMessage(te.getDescription(), -1, -1, te.getDuration());
    }

    /**
     * @return
     * @throws IOException
     */
    public String getPath() throws IOException {
        return log.getCanonicalPath();
    }

    /**
     * @return
     * @throws IOException
     */
    public Reader getReader() throws IOException {
        if (null != writer) {
            throw new IOException("writer is still open");
        }
        return new FileReader(log);
    }
}
